% Copyright (C) 2015 by Daniel Kraft <d@domob.eu>
% OctConf 2015 - Octave's profiler

%   CHI = floodfill_2 (IMG, A, B, R, C, MAXDIST = Inf)
%
% "Real" implementation of floodfill based on the correct algorithm.
% This is faster than 'floodfill' but still not optimal since it
% includes a lot of interpretation overhead from Octave.

function chi = floodfill_2 (img, a, b, r, c, maxDist = Inf)
  n = size (img, 1);
  assert (n, size (img, 2));

  dist = Inf (size (img));
  dist(r, c) = 0;
  assert (img(r, c) >= a && img(r, c) <= b);

  % We use a chunk of memory that is not adjusted for actual demand
  % and also where free'd elements are not recycled for the queue.
  % This makes the implementation easy and it should not be "too" bad
  % since it is comparable in size to chi and img.
  queue = NA (n^2, 2);
  queue(1, :) = [r, c];
  queueStart = 1;
  queueEnd = 2;

  % Perform classical floodfill, i. e., basically Dijkstra's algorithm
  % based on a grid.
  deltas = [-1, 0; 1, 0; 0, -1; 0, 1];
  while (queueStart < queueEnd)
    %assert (queueStart <= n^2 && queueEnd <= n^2);
    curR = queue(queueStart, 1);
    curC = queue(queueStart, 2);
    ++queueStart;
    %assert (isfinite (curR) && isfinite (curC) && dist(curR, curC) < maxDist);

    for i = 1 : size (deltas, 1)
      rp = curR + deltas(i, 1);
      cp = curC + deltas(i, 2);

      if (rp >= 1 && rp <= n && cp >= 1 && cp <= n && isinf (dist(rp, cp))
            && img(rp, cp) >= a && img(rp, cp) <= b)
        dist(rp, cp) = dist(curR, curC) + 1;

        if (dist(rp, cp) < maxDist)
          %assert (queueEnd < n^2);
          queue(queueEnd, :) = [rp, cp];
          ++queueEnd;
        endif
      endif
    endfor
  endwhile

  chi = isfinite (dist);
  %assert (chi, floodfill (img, a, b, r, c, maxDist));
endfunction
