/*
Copyright (C) 2015 by Daniel Kraft <d@domob.eu>
OctConf 2015 - Octave's profiler
*/

/* C++ implementation of floodfill.  */

#include <octave/oct.h>
#include <octave/boolMatrix.h>

#include <limits>
#include <queue>
#include <stdexcept>

struct Point
{

  int r;
  int c;
  int dist;

  inline Point (int rr, int cc, int d)
    : r(rr), c(cc), dist(d)
  {}

  inline Point (const Point& base, int dr, int dc)
    : r(base.r + dr), c(base.c + dc), dist(base.dist + 1)
  {}

};

typedef std::queue<Point> PointQueue;

DEFUN_DLD (floodfill_3, args, nargout,
  "  CHI = floodfill_3 (IMG, A, B, R, C, MAXDIST = Inf)\n\n"
  "Fast C++ implementation of floodfill.\n")
{
  try
    {
      if (args.length () < 5 || args.length () > 6 || nargout > 1)
        throw std::runtime_error ("invalid call to internalFloodFill");

      /* Get all arguments in.  */
      const Matrix img = args(0).matrix_value ();
      const double a = args(1).double_value ();
      const double b = args(2).double_value ();
      const int r = args(3).int_value () - 1;
      const int c = args(4).int_value () - 1;
      double maxDist = std::numeric_limits<double>::infinity ();
      if (args.length () >= 6)
        maxDist = args(5).double_value ();

      /* Extract image dimension and check for a square image
         (for simplicity).  */
      const dim_vector size = img.dims ();
      if (size.length () > 2)
        throw std::runtime_error ("expected at most two dimensions for IMG");
      assert (size.length () == 2);
      const int n = size(0);
      if (size(1) != n)
        throw std::runtime_error ("expected square image");

      /* Initialise things.  */
      boolMatrix chi(n, n);
      chi.fill (false);
      PointQueue queue;
      queue.push (Point (r, c, 0));

      /* Perform main loop.  */
      while (!queue.empty ())
        {
          const Point cur = queue.front ();
          queue.pop ();

          if (cur.r < 0 || cur.r >= n || cur.c < 0 || cur.c >= n)
            continue;
          if (cur.dist > maxDist || chi(cur.r, cur.c))
            continue;
          if (img(cur.r, cur.c) < a || img(cur.r, cur.c) > b)
            continue;

          chi(cur.r, cur.c) = true;
          queue.push (Point (cur, -1, 0));
          queue.push (Point (cur, 1, 0));
          queue.push (Point (cur, 0, -1));
          queue.push (Point (cur, 0, 1));
        }

      /* Return.  */
      octave_value_list res;
      res(0) = chi;

      return res;
    }
  catch (const std::runtime_error& exc)
    {
      error (exc.what ());
      return octave_value_list ();
    }
}
